#!/usr/bin/perl -w
#
# Rubi's Ultimate Backup Implementation
#
# Version 3.2, (C) 3.1999 by Beat Rubischon <beat@rubis.ch>
#                  8.1999 by Beat Rubischon <beat.rubischon@trivadis.com>
#                  2.2000 by Beat Rubischon <beat@rubis.ch>
#                  2.2004 by Beat Rubischon <rubischon@phys.ethz.ch>
#                  7.2007 by Beat Rubischon <beat@0x1b.ch>
#                 10.2010 by Beat Rubischon <beat.rubischon@dalco.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# $Id: backup.pl,v 1.16 2010-10-14 11:45:40 beat Exp $
#
# Main handler for backup

use strict;

my $PREFIX=$ARGV[0];
my $SSH=$ARGV[1];
my $MKFIFO=$ARGV[2];
my $CPIO=$ARGV[3];
my $RSYNC=$ARGV[4];
my $SRCCPIO=$ARGV[5];
my $SRCRSYNC=$ARGV[6];
my $SRCHOST=$ARGV[7];
my $SRCPART=$ARGV[8];
my $BASEDST=$ARGV[9];
my $DSTDIR=$ARGV[10];
my $OLDDST=$ARGV[11];
my $UTIME=$ARGV[12];

my $retcd=1;

my $fccount=0; my $fcbytes=0;
my $flcount=0; my $flbytes=0;
my $scount=0;
my $dcount=0;
my $fcount=0;
my $ccount=0;
my $bcount=0;
my $kcount=0;
my $ucount=0;

my ($walkcmd, $cpiocmd);

$|=1;

my $time=`date`;

print "* * * Backup report * * *\n";
print "----------------------------------------------------------------------------\n";
print "Backing up:             $SRCHOST:$SRCPART\n";
print "        to:             $DSTDIR\n";
print "last backup in:         ";

if ($OLDDST eq "") {
    print "no previus backup done\n";
} else {
    print "$OLDDST\n";
}

if ($UTIME) {
  open UTIME, ">".$BASEDST."/utimes";
}

print "Start time:             $time\n";

if ($SRCHOST eq "localhost") {
    $walkcmd=$PREFIX."/lib/walk.pl ".$SRCPART."|";
    $cpiocmd="|".$CPIO." -p -u -m -0 ".$DSTDIR;
} else {
    $walkcmd=$SSH." ".$SRCHOST." ".$PREFIX."/lib/walk.pl ".$SRCPART."|";
    $cpiocmd="|".$SSH." ".$SRCHOST." ".$SRCCPIO." -o -H newc -0 | ".$CPIO." -i -d -u -m --no-absolute-filenames";
}

open(WALK, $walkcmd) || die "cant open $walkcmd";
open(CPIO, $cpiocmd) || die "cant open $cpiocmd";

while(<WALK>) {
    chomp;

    if (/\#\#\#EOF\#\#\#/) {
	$retcd=0;
	last;
    }

    my ($type, $rdev, $rino, $rmode, $rnlink, $ruid, $rgid, $rrdev,
      $rsize, $ratime, $rmtime, $rctime, $rblksize, $rblocks,
      $file) = (m/^([-0-9]*)\/([-0-9]*)\/([-0-9]*)\/([-0-9]*)\/([-0-9]*)\/([-0-9]*)\/([-0-9]*)\/([-0-9]*)\/([-0-9]*)\/([-0-9]*)\/([-0-9]*)\/([-0-9]*)\/([-0-9]*)\/([-0-9]*)(\/.*)$/o);

    $file=~tr/\000/\n/;

    # regular file
    if ($type eq "10") {
	my ($dev, $ino, $mode, $nlink, $uid, $gid, $rdev,
	  $size, $atime, $mtime, $ctime, $blksize,
	  $blocks);

	# > 2GB, copy with rsync
	if ($rsize > 2147483647) {
	    if ($RSYNC eq "" or $SRCRSYNC eq "") {
		print "File $file more then 2GBytes and no rsync available!\n";
	    } else {
	        if ($OLDDST ne "") {
		    ($dev, $ino, $mode, $nlink, $uid, $gid, $rdev, 
	  	      $size, $atime, $mtime, $ctime, $blksize,
	              $blocks) = lstat($OLDDST.$file);
		    link $OLDDST.$file,$DSTDIR.$file;
	        }

	        if ($OLDDST eq "" or ! defined $dev or $mtime ne $rmtime) {
		    my @CMD;
		    if ($SRCHOST eq "localhost") {
			@CMD=($RSYNC, '-av', $file, $DSTDIR.$file);
		    } else {
			(my $file_esc = $file) =~ s/\'/\'\"\'\"\'/gs;
			@CMD=($RSYNC, '-e', $SSH, '--rsync-path='.$SRCRSYNC,
			  '-av', $SRCHOST.":'$file_esc'", $DSTDIR.$file);
		    }

		    system(@CMD);

		    $fccount++;
		    $fcbytes=$fcbytes+$rsize;
		} else {
		    $flcount++;
		    $flbytes=$flbytes+$rsize;
		}
	    }

	# < 2GB, copy with cpio
	} else {
	    if ($OLDDST ne "") {
		($dev, $ino, $mode, $nlink, $uid, $gid, $rdev, 
	  	  $size, $atime, $mtime, $ctime, $blksize,
	          $blocks) = lstat($OLDDST.$file);
	    }

	    if ($OLDDST eq "" or ! defined $dev or $mtime ne $rmtime) {
	        print CPIO "$file\000";
	        $fccount++;
	        $fcbytes=$fcbytes+$rsize;
	    } else {
	        link $OLDDST.$file,$DSTDIR.$file;
	        $flcount++;
	        $flbytes=$flbytes+$rsize;
	    }
	}
    }

    # symlink
    elsif($type eq "12") {
	print CPIO "$file\000";
	$scount++;
    }

    # directory
    elsif ($type eq "04") {
	my @dirs=split(/\//,$file);
	my $base=$DSTDIR;

	foreach(@dirs) {
	    if ($_ ne "") {
		$base=$base."/".$_;
		if (stat($base) eq "") {
                    if ($base =~ m/\n/s) {
                        warn "Directory '$base' contains newline";
                    }
		    mkdir($base, 0755);
		}
	    }
        }

	chmod $rmode, $DSTDIR.$file;
	chown $ruid, $rgid, $DSTDIR.$file;
	if ($UTIME) {
	  my $tempfile=$file;
	  $tempfile=~tr/\n/\000/;
	  print UTIME "$ratime/$rmtime/$tempfile\n";
	}
	$dcount++;  
    }

    # fifo
    elsif ($type eq "01") {
	unless (fork) {
	    exec($MKFIFO." ".$DSTDIR.$file);
        }
	chmod $rmode, $DSTDIR.$file;
	chown $ruid, $rgid, $DSTDIR.$file;
	if ($UTIME) {
	  my $tempfile=$file;
	  $tempfile=~tr/\n/\000/;
	  print UTIME "$ratime/$rmtime/$tempfile\n";
	}
	$fcount++;
    }
  
    # char special
    elsif ($type eq "02") {
	print CPIO "$file\000";
	$ccount++
    }
  
    # block special
    elsif ($type eq "06") {
	print CPIO "$file\000";
	$bcount++;
    }
  
    # socket
    elsif ($type eq "14") {
	$kcount++;
    }
  
    # unknown
    else {
	print STDERR "$file has unknown type $type.\n";
	$ucount++;
    }
}

close(WALK);
close(CPIO);

if ($UTIME) {
  close(UTIME);

  print STDERR "\nSetting atime / utime...";

  open UTIME, "<".$BASEDST."/utimes";

  while(<UTIME>) {
    my ($ratime, $rmtime, $file) = 
      (m/^([-0-9]*)\/([-0-9]*)\/(\/.*)$/o);
    $file=~tr/\000/\n/;

    utime $ratime, $rmtime, $DSTDIR.$file;
  }

  close UTIME;

  print "done.\n";
}

my $fcval="B";
if ($fcbytes >= 1024) {
    $fcbytes=$fcbytes / 1024;
    $fcval="K";
    if ($fcbytes >= 1024) {
	$fcbytes=$fcbytes / 1024;
	$fcval="M";
	if ($fcbytes >= 1024) {
	    $fcbytes=$fcbytes / 1024;
	    $fcval="G";
	}
    }
}

my $flval="B";
if ($flbytes >= 1024) {
    $flbytes=$flbytes / 1024;
    $flval="K";
    if ($flbytes >= 1024) {
	$flbytes=$flbytes / 1024;
	$flval="M";
	if ($flbytes >= 1024) {
	    $flbytes=$flbytes / 1024;
	    $flval="G";
	}
    }
}

if ($retcd) {
    print "\nNo EOF from remote host!\n";
}

$time=`date`;
print "\nEnd time:               $time\n";
print "Type                    Count     Size\n";
print "----------------------------------------------------------------------------\n";
printf "%-22s %6i %6.1f %s\n", "files copied:", $fccount, $fcbytes, $fcval;
printf "%-22s %6i %6.1f %s\n", "files linked:", $flcount, $flbytes, $flval;
printf "%-22s %6i\n", "directories:", $dcount;
printf "%-22s %6i\n", "symlinks:", $scount;
printf "%-22s %6i\n", "char special:", $ccount;
printf "%-22s %6i\n", "block special:", $bcount;
printf "%-22s %6i\n", "fifos:", $fcount;
printf "%-22s %6i not copied\n", "sockets:", $kcount;
printf "%-22s %6i not copied\n", "unknown:", $ucount;
print "----------------------------------------------------------------------------\n";
print "\n";

exit $retcd;
